
import java.awt.*;
import java.awt.image.*;
import java.io.*;

// Given a list of images on the command-line, create 1-segment
// segmentation files in the specified directory.
public class CreateEmptySeg {

    // Preconditions: None.
    // Actions: Wait until the given image has completely loaded.
    // Return Value: true on any error; false on success.
    static boolean waitForImage (Image image) {
	MediaTracker tracker = new MediaTracker (new Container ());
	try {
	    tracker.addImage (image, 0);
	    tracker.waitForID (0);
	}
	catch (InterruptedException e) {
	    return true;
	}
	return tracker.isErrorID (0);
    }

    // Preconditions: None.
    // Actions: Load the named image.  
    // Return Value: The image on success; null on any error.
    static Image loadImage (File file) {
	String imageFileName = file.getAbsolutePath();
	Image image = Toolkit.getDefaultToolkit().createImage (imageFileName);
	boolean error = waitForImage (image);
	return error ? null : image;
    }

    static int userID = 0;
    static String segDir = ".";

    // Print usage and exit.
    static void usage () {
	System.err.println (
	    "\n" + 
	    "usage: CreateEmptySeg [options] image1 image2 ...\n" +
	    "options:\n" + 
	    "  -segdir <dir>   Set directory for .seg files.\n");
	System.exit (1);
    }

    static int processArgs (String args[]) {
	int i;
	for (i = 0; i < args.length; i++) {
	    if (args[i].equals ("-segdir")) {
		if (++i == args.length) { usage (); }
		segDir = args[i];
	    } else {
		break;
	    }
	}
	if (i == args.length) { usage (); }
	return i;
    }

    public static void main (String[] args) {

	int i = processArgs (args);

	for (; i < args.length; i++) {
	    File imageFile = new File (args[i]);
	    String imageFileName = imageFile.getName ();

	    // Load the image.
	    Image image = loadImage (imageFile);
	    if (image == null) {
		System.err.println ("Error opening image file '" + args[i] + "'.");
		continue;
	    }

	    // Get the image width/height.
	    int imWidth = image.getWidth (null);
	    int imHeight = image.getHeight (null);

	    // Compute the segmentation filename.
	    int index = imageFileName.lastIndexOf ('.');
	    if (index == -1) {
		index = imageFileName.length();
	    }
	    String imageFileNameRoot = imageFileName.substring (0, index);
	    String segFileName = segDir + "/" + imageFileNameRoot + ".seg";
	    System.out.println ("Writing file '" + segFileName + "'...");

            // Write the segmentation file.
	    FileWriter writer = null;
	    try {
		File segFile = new File (segFileName);
		writer = new FileWriter (segFile);
		writer.write ("image " + imageFileNameRoot + "\n");
		writer.write ("width " + imWidth + "\n");
		writer.write ("height " + imHeight + "\n");
		writer.write ("segments 1\n");
		writer.write ("data\n"); 
		int col = imWidth - 1;
		for (int row = 0; row < imHeight; row++) {
		    writer.write ("0 " + row + " 0 " + col + "\n");
		}
	    } catch (IOException e) {
		System.err.println ("Error writing segmentation file '" + segFileName + "' for image '" + args[i] + "'.");
	    } finally {
		if (writer != null) { 
		    try { writer.close (); } catch (IOException e) {}
		}
	    }
	}

	System.exit (0);
    }
}
