
// Author:  David Martin  <dmartin@eecs.berkeley.edu>

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.util.*;
import java.io.*;
import java.net.*;
import gnu.regexp.*;
import java.util.List;

//
// This application has two operating modes: client mode and hermit mode.
//
// In client mode, most all operations are done in conjuction with
// an image server.  In this mode, the 'user' and 'image' fields in
// the segmentation files must be numeric, corresponding to user
// and image IDs.  The superUser flag applies to this mode only, so
// that we can present a simpler interface to naive users.
//
// In hermit mode, there is no server.  All image/segfile operations
// are done either on local files or over HTTP.  In this mode, the
// 'user' field is not used at all, and the 'image' field is used
// simply for documentation (the image file name is placed there).
//

public class SegmentApp
{
    final static int CLIENT_MODE = 1;
    final static int HERMIT_MODE = 2;

    final static int defaultMode = CLIENT_MODE;
    static int mode = defaultMode;

    static boolean superUser = false;

    static String modeToString (int mode) {
	switch (mode) {
	    case CLIENT_MODE: return "client";
	    case HERMIT_MODE: return "hermit";
	    default: throw new InternalError ("Illegal mode: " + mode);
	}
    }

    // If this flag is true, images are displayed in grayscale.
    final static boolean defaultGrayImage = true;
    static boolean grayImage = defaultGrayImage;

    final static boolean defaultInvertImage = false;
    static boolean invertImage = defaultInvertImage;

    final static boolean defaultFlipImage = false;
    static boolean flipImage = defaultFlipImage;

    // The place we look for images.
    final static String defaultImageDirLoc = "http://www.cs.berkeley.edu/~dmartin/segment/images/medium";
    static String imageDirLoc = defaultImageDirLoc;

    // Do we allow users to open segmentation files for other users?
    static boolean forceUserIDMatch = true;

    // Magnification limits.
    final static double minMagnification = 0.1;
    final static double maxMagnification = 4.0;

    // The image.
    final static int INVALID_IMAGE_ID = -1;
    static int imageID = INVALID_IMAGE_ID;
    static Image originalImage = null;		// The original image, as loaded (null if none loaded).
    static int[] originalPixels = null;		// The ARGB pixels of the original image.
    static BufferedImage displayImage = null;	// The image as we display it.
    static int imWidth = -1;
    static int imHeight = -1;	

    // The user.
    final static int INVALID_USER_ID = -1;
    static int userID = INVALID_USER_ID;
    static String userName;
    static int totalImageCount = 0;	// How many images this user has segmented total.
    static int sessionImageCount = 0;	// How many images this user has segmented this session.

    // The segmentation and selection.
    static Segmentation segmentation = null;
    static Selection selection = null;

    // This flag is true when there are unsaved changes.
    static boolean unsavedChanges = false;

    static MainWindow mainWindow;

    // The image ID server.
    static final String defaultImageIDServerHost = "ribbit.cs.berkeley.edu";
    static final int defaultImageIDServerPort = 2859;
    static String imageIDServerHost = defaultImageIDServerHost;
    static int imageIDServerPort = defaultImageIDServerPort;

    static String userSegDir = "Segmentations";
    
    // Filenames for current segment/image.
    // These are valid in both modes, but used only in hermit mode.
    static String segFileName;		// Whole path.
    static String imageFileName;	// File part only.

    // Width of help/info dialogs.
    static final int helpWidth = 120;
    static final int infoWidth = 50;

    // Is an image loaded?
    static boolean imageLoaded () {
	return (originalImage != null);
    }

    // Return List of Integer image IDs.
    static List getSubmittedSegList () throws Exception {
	Util.assert (mode == CLIENT_MODE);
	SegServer server = null;
	try {
	    // Connect to server.
	    server = new SegServer (imageIDServerHost, imageIDServerPort);
	    server.connect ();

	    // Send request.
            server.sendLine ("get seg list");
	    server.sendLine ("userID " + userID);
	    server.sendDone ();

	    // Process reply.
	    Map reply = server.getReply ();
	    if (reply == null) { throw new Exception ("Server error."); }
	    List replyList = (List) reply.get ("seglist");
	    List idList = new ArrayList ();
	    for (int i = 0; i < replyList.size(); i++) {
		int id = Integer.parseInt ((String)replyList.get (i));
		idList.add (new Integer (id));
	    }
	    return idList;

	} finally {
	    if (server != null) { server.disconnect (); } 
	}
    }

    public static class ImageRequest {
	public boolean allDone;
	public int imageID;
	public boolean gray, invert, flipflop;
	public ImageRequest (boolean allDone, int imageID, 
			     boolean gray, boolean invert, boolean flipflop) {
	    this.allDone = allDone;
	    this.imageID = imageID;
	    this.gray = gray;
	    this.invert = invert;
	    this.flipflop = flipflop;
	}
    }

    // Return new image ID, selected by the server.
    static ImageRequest getNewImageID () throws Exception {
	Util.assert (mode == CLIENT_MODE);
	SegServer server = null;
	try {
	    // Connect to server.
	    server = new SegServer (imageIDServerHost, imageIDServerPort);
	    server.connect ();

	    // Send request.
            server.sendLine ("new image");
	    server.sendLine ("userID " + userID + "");
	    server.sendDone ();

	    // Process reply.
	    Map reply = server.getReply ();
	    if (reply == null) { throw new Exception ("Server error."); }
	    if (Integer.parseInt ((String) reply.get ("alldone")) == 1) {
		return new ImageRequest (true, 0, false, false, false);
	    } 
	    int imageID = Integer.parseInt ((String) reply.get ("imageID"));
	    boolean gray = (Integer.parseInt ((String) reply.get ("gray")) == 1);
	    boolean invert = (Integer.parseInt ((String) reply.get ("invert")) == 1);
	    boolean flipflop = (Integer.parseInt ((String) reply.get ("flipflop")) == 1);
	    return new ImageRequest (false, imageID, gray, invert, flipflop);

	} finally {
	    if (server != null) { server.disconnect (); } 
	}
    }

    static Image loadImageByID (int id) throws Exception {
	URL url = new URL (imageDirLoc + "/" + id + ".jpg");
	return Util.loadImage (url);
    }

    // Print usage and exit.
    static void usage () {
	System.err.print (
	    "usage: SegmentApp [options]\n" +
	    "options [all modes]:\n" +
	    "  -mode <mode>     Set application mode.  One of:\n" +
	    "                     " + modeToString(CLIENT_MODE) + (defaultMode == CLIENT_MODE ? " [default]" : "") + "\n" +
	    "                     " + modeToString(HERMIT_MODE) + (defaultMode == HERMIT_MODE ? " [default]" : "") + "\n" +
	    "  -gray            Convert image to grayscale" + (defaultGrayImage ? " [default]" : "") + ".\n" +
	    "  -flip            Flip image vertically" + (defaultFlipImage ? " [default]" : "") + ".\n" +
	    "  -invert          Invert pixel values" + (defaultInvertImage ? " [default]" : "") + ".\n" +
	    "options [client mode]:\n" +
	    "  -user <id>       Set user ID (non-negative integer).\n" +
	    "  -server <s>      Set server.\n" +
	    "                   [default " + defaultImageIDServerHost + "]\n" +
	    "  -port <n>        Set server port [default " + defaultImageIDServerPort + "].\n" +
	    "  -imdir <url>     Set directory for images.\n" +
	    "                   [default " + defaultImageDirLoc + "]\n" +
	    "");
	System.exit (1);
    }

    static void processArgs (String args[]) {
	for (int i = 0; i < args.length; i++) {
	    if (args[i].equals ("-su")) {
		superUser = true;
	    } else if (args[i].equals ("-mode")) {
		if (++i == args.length) { usage (); }
		if (args[i].equals (modeToString (CLIENT_MODE))) {
		    mode = CLIENT_MODE;
		} else if (args[i].equals (modeToString (HERMIT_MODE))) {
		    mode = HERMIT_MODE;
		} else {
		    usage ();
		}
	    } else if (mode == CLIENT_MODE && args[i].equals ("-user")) {
		if (++i == args.length) { usage (); }
		try {
		    userID = Integer.decode(args[i]).intValue();
		} catch (NumberFormatException e) {
		    userID = INVALID_USER_ID;
		}
		if (userID < 0) { usage (); }
	    } else if (mode == CLIENT_MODE && args[i].equals ("-imdir")) {
		if (++i == args.length) { usage (); }
		imageDirLoc = args[i];
	    } else if (args[i].equals ("-gray")) {
		grayImage = true;
	    } else if (args[i].equals ("-flip")) {
		flipImage = true;
	    } else if (args[i].equals ("-invert")) {
		invertImage = true;
	    } else if (mode == CLIENT_MODE && args[i].equals ("-server")) {
		if (++i == args.length) { usage (); }
		imageIDServerHost = args[i];
	    } else if (mode == CLIENT_MODE && args[i].equals ("-port")) {
		if (++i == args.length) { usage (); }
		try {
		    imageIDServerPort = Integer.decode(args[i]).intValue();
		} catch (NumberFormatException e) {
		    imageIDServerPort = -1;
		}
		if (imageIDServerPort < 0) { usage (); }
	    } else {
		usage ();
	    }
	}
    }

    // Return INVALID_USER_ID if user presses cancel.
    // Return non-negative integer id.
    static int promptUserForID () {
	int id = INVALID_USER_ID;
	while (id == INVALID_USER_ID) {
	    String s = JOptionPane.showInputDialog (
		null, "Enter your user ID.",
		"Set User ID", JOptionPane.INFORMATION_MESSAGE);
	    if (s == null) {
		// User pressed cancel button in InputDialog.
		return INVALID_USER_ID;
	    }
	    try {
		id = Integer.decode(s).intValue();
	    } catch (NumberFormatException e) {
		id = INVALID_USER_ID;
	    }
	    if (id == INVALID_USER_ID || id < 0) {
		JOptionPane.showMessageDialog (
		    null, Util.lineBreak (
			"The ID you entered is not valid. " +
			"Please enter a non-negative integer ID. " +
			"", infoWidth),
		    "Error", JOptionPane.ERROR_MESSAGE);
		id = INVALID_USER_ID;
	    }
	}
	return id;
    }

    static void changeUserID () {
	int uid = promptUserForID ();
	if (uid < 0) {
	    // User pressed cancel button in InputDialog.
	    return;
	}
	checkUserID (uid, false);
    }

    // Query server for information about userID.
    // Return true if user ID is ok.
    // Sets userName/totalImageCount as side-effect (I know: yuk).
    static boolean checkUserID (int userID, boolean confirm) {
	Util.assert (mode == CLIENT_MODE);
	SegServer server = null;
	try {
	    // Connect to server.
	    server = new SegServer (imageIDServerHost, imageIDServerPort);
	    server.connect ();

	    // Send request.
            server.sendLine ("lookup user");
	    server.sendLine ("userID " + userID);
	    server.sendDone ();

	    // Process reply.
	    Map reply = server.getReply ();
	    if (reply == null) { throw new Exception ("Server error."); }
	    int found = Integer.parseInt ((String) reply.get ("found"));
	    if (found == 0) {
		JOptionPane.showMessageDialog (
		    null, Util.lineBreak (
			"The user ID \"" + userID + "\" is not valid." +
			"", infoWidth),
		    "Error", JOptionPane.ERROR_MESSAGE);
		return false;
	    }
	    String name = (String) reply.get ("name");
	    String email = (String) reply.get ("email");
	    int segCount = Integer.parseInt ((String) reply.get ("segcount"));

	    int answer;
	    if (confirm) {
		// Ask user to confirm identity.
		answer = JOptionPane.showConfirmDialog (
		    mainWindow, 
		    "Are you " + name + " (" + email + ")?",
		    "Question", JOptionPane.YES_NO_OPTION);
	    } else {
		answer = JOptionPane.YES_OPTION;
	    }
	    if (answer == JOptionPane.YES_OPTION) {
		SegmentApp.userID = userID;
		userName = name;
		totalImageCount = segCount;
		return true;
	    } else {
		return false;
	    }
	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		null, Util.lineBreak (
		    "The server may be down.  Error: " + 
		    e.getMessage() + 
		    "", infoWidth),
		"Error", JOptionPane.ERROR_MESSAGE);
	    return false;
	} finally {
	    if (server != null) { server.disconnect (); } 
	}
    }
    
    static int raffleTickets (int images) {
	if (images < 6) { return 0; }
	if (images < 21) { return images - 5; }
	return 15 + 2 * (images - 20);
    }

    static void main (String args[]) {
	processArgs (args);

	if (mode == CLIENT_MODE) {
	    // Get a valid user ID from user.
	    while (true) {
		int uid = (userID >= 0) ? userID : promptUserForID ();
		if (uid < 0) {
		    // User pressed cancel button in InputDialog.
		    System.exit (0);
		}
		boolean ok = checkUserID (uid, true);
		if (ok) { break; }
		userID = INVALID_USER_ID;
	    }
	} else {
	    userID = INVALID_USER_ID;
	    superUser = true;
	}

	// Define user segmentation directory, and make sure the
	// directory exists.
	String homeDirPath = System.getProperty ("user.home");
	if (homeDirPath != null) {
	    userSegDir = homeDirPath.concat ("/" + userSegDir);
	}
	if (mode == CLIENT_MODE) {
	    userSegDir = userSegDir.concat ("/" + userID);
	}
	File userDir = new File (userSegDir);
	if (!userDir.exists()) {
	    boolean made = userDir.mkdirs ();
	    if (!made) {
		System.err.println (
		    "Could not create directory '" + userDir + "'.");
		System.exit (1);
	    }
	}

	mainWindow = new MainWindow ("Segmentation Tool");
	mainWindow.show ();
    }

    static void clearImage () {
	imageID = INVALID_USER_ID;
	originalImage = null;
	imWidth = imHeight = -1;
	segmentation = null;
	originalPixels = null;
	displayImage = null;
	selection = null;
	undoActions = new ArrayList ();
	redoActions = new ArrayList ();
	mainWindow.splitterWindow.clearSegments();
	unsavedChanges = false;
    }

    // Keep current image; install new segmentation.
    static void installNewSegmentation (Segmentation newSeg, 
					String newSegFileName) throws Exception {
	if (mode == CLIENT_MODE && forceUserIDMatch) {
	    // Make sure user IDs match.
	    // Allow any user to open files that don't specify
	    // a user id.
	    if (newSeg.user == null) { newSeg.user = "" + userID; }
	    int newUserID = newSeg.getUserID();
	    if (newUserID != userID) {
		throw new Exception (
		    "User ID in new segmentation (" + newUserID 
		    + ") does not match the current user ID (" 
		    + userID + ").");
	    }
	}

	if (mode == CLIENT_MODE) {
	    // Make sure image IDs match.
	    if (newSeg.image == null) { newSeg.image = "" + imageID; }
	    int newImageID = newSeg.getImageID();
	    if (newImageID != imageID) {
		throw new Exception (
		    "Image ID in segmentation (" + newImageID
		    + ") does not match the current image ID ("
		    + imageID + ").");
	    }
	}

	// Make sure the width/height match.
	int width = imWidth;
	int height = imHeight;
	if (newSeg != null 
	    && (width != newSeg.width || height != newSeg.height)) {
	    throw new Exception (
		"The image dimensions (" + width + "x" + height  
		+ ") do not match the segmentation dimensions (" 
		+ newSeg.width + "x" + newSeg.height + ").");
	}

	grayImage = newSeg.gray;
	invertImage = newSeg.invert;
	flipImage = newSeg.flipflop;

	segmentation = newSeg;
	if (flipImage) { segmentation.doFlipFlop (); }

	segFileName = newSegFileName;
	if (mode == HERMIT_MODE) {
	    segmentation.image = imageFileName;
	}

	computeDisplayImage ();

	// Setup the selection.
	Util.assert (segmentation.nonEmptySegmentCount() == segmentation.newSegmentID());
	selection = new Selection (segmentation.nonEmptySegmentCount());

	// Clear undo/redo history.
	undoActions = new ArrayList ();
	redoActions = new ArrayList ();

	mainWindow.newImage ();
	unsavedChanges = false;
    }

    // Install new image and fresh segmentation.
    static void installNewImage (int newImageID, boolean gray, 
				 boolean invert, boolean flipflop,
				 Image newImage,
				 String newImageFileName) throws Exception {
	// Get the pixels of the original image.
	int width = newImage.getWidth (null);
	int height = newImage.getHeight (null);
	originalPixels = new int [width * height];
	PixelGrabber pg = new PixelGrabber (
	    newImage,
	    0, 0, width, height,
	    originalPixels, 0, width);
	pg.grabPixels ();
	Util.assert ((pg.getStatus() & ImageObserver.ALLBITS) != 0);

	imageID = newImageID;
	originalImage = newImage;
	imWidth = width;
	imHeight = height;
	imageFileName = newImageFileName;

	// Allocate space for the display image.
	displayImage = new BufferedImage (
	    imWidth, imHeight, BufferedImage.TYPE_INT_ARGB);

	SegmentApp.grayImage = gray;
	SegmentApp.invertImage = invert;
	SegmentApp.flipImage = flipflop;

	computeDisplayImage ();

	if (mode == CLIENT_MODE) {
	    segmentation = new Segmentation (
		imWidth, imHeight, 0, "" + newImageID, "" + userID, 
		grayImage, invertImage, flipImage);
	} else {
	    segmentation = new Segmentation (
		imWidth, imHeight, 0, imageFileName, null, 
		grayImage, invertImage, flipImage);
	}

	// Setup the selection.
	Util.assert (segmentation.nonEmptySegmentCount() == segmentation.newSegmentID());
	selection = new Selection (segmentation.nonEmptySegmentCount());

	// Clear undo/redo history.
	undoActions = new ArrayList ();
	redoActions = new ArrayList ();

	mainWindow.newImage ();
	unsavedChanges = false;
    }

    // Install new image and new segmentation simultaneously.
    static void installNewImageAndSegmentation (Image newImage, 
						Segmentation newSeg,
						String newImageFileName,
						String newSegFileName) throws Exception {
	if (mode == CLIENT_MODE && forceUserIDMatch) {
	    // Make sure user IDs match.
	    // Allow any user to open files that don't specify
	    // a user id.
	    if (newSeg.user == null) { newSeg.user = "" + userID; }
	    int newUserID = newSeg.getUserID();
	    if (newUserID != userID) {
		throw new Exception (
		    "User ID in new segmentation (" + newUserID 
		    + ") does not match the current user ID (" 
		    + userID + ").");
	    }
	}

	// Make sure the width/height match.
	int width = newImage.getWidth (null);
	int height = newImage.getHeight (null);
	if (newSeg != null 
	    && (width != newSeg.width || height != newSeg.height)) {
	    throw new Exception (
		"The image dimensions (" + width + "x" + height  
		+ ") do not match the segmentation dimensions (" 
		+ newSeg.width + "x" + newSeg.height + ").");
	}

	int newImageID = newSeg.getImageID();

	// Get the pixels of the original image.
	originalPixels = new int [width * height];
	PixelGrabber pg = new PixelGrabber (
	    newImage, 
	    0, 0, width, height, 
	    originalPixels, 0, width);
	pg.grabPixels ();
	Util.assert ((pg.getStatus() & ImageObserver.ALLBITS) != 0);

	imageID = newImageID;
	originalImage = newImage;
	imWidth = width;
	imHeight = height;
	imageFileName = newImageFileName;
	segFileName = newSegFileName;

	// Allocate space for the display image.
	displayImage = new BufferedImage (
	    imWidth, imHeight, BufferedImage.TYPE_INT_ARGB);

	grayImage = newSeg.gray;
	invertImage = newSeg.invert;
	flipImage = newSeg.flipflop;

	segmentation = newSeg;
	if (flipImage) { segmentation.doFlipFlop (); }

	computeDisplayImage ();

	if (mode == HERMIT_MODE) {
	    segmentation.image = imageFileName;
	}

	// Setup the selection.
	Util.assert (segmentation.nonEmptySegmentCount() == segmentation.newSegmentID());
	selection = new Selection (segmentation.nonEmptySegmentCount());

	// Clear undo/redo history.
	undoActions = new ArrayList ();
	redoActions = new ArrayList ();

	mainWindow.newImage ();
	unsavedChanges = false;
    }

    // Apply image transformations, if necessary.
    static void computeDisplayImage () {
	for (int y = 0; y < imHeight; y++) {
	    for (int x = 0; x < imWidth; x++) {
		int rgb = originalPixels [y * imWidth + x];
		if (grayImage) { rgb = RGBFilter.toGray (rgb); }
		if (invertImage) { rgb = RGBFilter.invert (rgb); }
		int newX = x;
		int newY = y;
		if (flipImage) { 
		    newX = imWidth - x - 1; 
		    newY = imHeight - y - 1; 
		}
		displayImage.setRGB (newX, newY, rgb);
	    }
	}
	//displayImage.setRGB (
	//0, 0, imWidth, imHeight,
	//originalPixels, 0, imWidth);
    }

    // Save the current segmentation to the file.
    static void saveSegmentation (File file) throws Exception {
	Util.assert (imageLoaded());
	FileWriter fileWriter = null;
	if (flipImage) { segmentation.doFlipFlop (); }
	try {
	    Reader segReader = segmentation.getReader ();
	    fileWriter = new FileWriter (file);
	    while (true) {
		int c = segReader.read ();
		if (c == -1) {
		    break;
		}
		fileWriter.write (c);
	    }
	    unsavedChanges = false;
	} finally {
	    if (flipImage) { segmentation.doFlipFlop (); }
	    try {
		if (fileWriter != null) { fileWriter.close (); } 
	    } catch (Exception e) {}
	}
    }

    static Segmentation openSegmentationByID (int imageID) throws Exception {
	Util.assert (mode == CLIENT_MODE);
	SegServer server = null;
	try {
	    // Connect to server.
	    server = new SegServer (imageIDServerHost, imageIDServerPort);
	    server.connect ();

	    // Send request.
            server.sendLine ("get seg");
	    server.sendLine ("userID " + userID);
	    server.sendLine ("imageID " + imageID);
	    server.sendDone ();

	    // Process reply.
	    Map reply = server.getReply ();
	    if (reply == null) { throw new Exception ("Server error."); }
	    List segData = (List) reply.get ("segdata");
	    String[] strings = new String [segData.size()];
	    segData.toArray (strings);
	    return new Segmentation (new BufferedStringReader (strings));
	} finally {
	    if (server != null) { server.disconnect (); } 
	}
    }

    static void abortSegmentation () throws Exception {
	Util.assert (mode == CLIENT_MODE);
	Util.assert (imageLoaded());
	SegServer server = null;
	try {
	    // Connect to server.
	    server = new SegServer (imageIDServerHost, imageIDServerPort);
	    server.connect ();

	    // Send request.
            server.sendLine ("abort seg");
	    server.sendLine ("userID " + userID);
	    server.sendLine ("imageID " + imageID);
	    server.sendDone ();

	    // Process reply.
	    Map reply = server.getReply ();
	    if (reply == null) { throw new Exception ("Server error."); }
	    clearImage ();

	} finally {
	    if (server != null) { server.disconnect (); } 
	}
    }

    static void submitSegmentation () throws Exception {
	Util.assert (mode == CLIENT_MODE);
	Util.assert (imageLoaded());
	SegServer server = null;
	if (flipImage) { segmentation.doFlipFlop (); }
	try {
	    // Set up streams for segmentation data.
	    Reader segReader = segmentation.getReader ();
	    BufferedReader bufSegReader = new BufferedReader (segReader);

	    // Connect to server.
	    server = new SegServer (imageIDServerHost, imageIDServerPort);
	    server.connect ();

	    // Send data to server.
            server.sendLine ("commit seg");
	    server.sendLine ("userID " + userID);
	    server.sendLine ("imageID " + imageID);
	    server.sendLine ("gray " + (grayImage ? 1 : 0));
	    server.sendLine ("invert " + (invertImage ? 1 : 0));
	    server.sendLine ("flipflop " + (flipImage ? 1 : 0));
	    server.sendLine ("begin segdata");
	    while (true) {
		String line = bufSegReader.readLine ();
		if (line == null) { break; }
		server.sendLine ("-" + line);
	    }
	    server.sendLine ("end");
	    server.sendDone ();

	    // Process reply.
	    Map reply = server.getReply ();
	    if (reply == null) { throw new Exception ("Server error."); }
	    clearImage ();
	    totalImageCount = Integer.parseInt((String)reply.get("segcount"));
	    sessionImageCount++;

	} finally {
	    if (flipImage && segmentation != null) { segmentation.doFlipFlop (); }
	    if (server != null) { server.disconnect (); } 
	}
    }

    // Calls exit if we quit; returns otherwise.
    static void quit () {
	if (unsavedChanges) {
	    int answer = JOptionPane.showConfirmDialog (
		mainWindow, Util.lineBreak (
		    "There are unsaved changes!  " +
		    "Are you sure you want to quit? " +
		    "", infoWidth),
		"Question", JOptionPane.YES_NO_OPTION);
	    if (answer == JOptionPane.YES_OPTION) {
		System.exit (0);
	    }
	} else if (mode == CLIENT_MODE && imageLoaded()) {
	    int answer = JOptionPane.showConfirmDialog (
		mainWindow, Util.lineBreak (
		    "Are you sure you want to quit without " +
		    "first submitting or aborting the current " +
		    "segmentation? " +
		    "", infoWidth),
		"Question", JOptionPane.YES_NO_OPTION);
	    if (answer == JOptionPane.YES_OPTION) {
		System.exit (0);
	    }
	} else {
	    System.exit (0);
	}
    }

    // A class to manage the selection of segments.
    static class Selection {
	Util.BoolArray selected;
	Selection (int count) {
	    selected = new Util.BoolArray ();
	}
	boolean selected (int id) {
	    return selected.get (id);
	}
	void selectAll (boolean val) {
	    for (int i = 0; i < selected.length(); i++) {
		selected.set (i, val);
	    }
	}
	void toggleSelection (int id) {
	    selected.set (id, !selected.get (id));
	}
	void deselectEmptySegments () {
	    for (int i = 0; i < selected.length(); i++) {
		if (selected (i)) {
		    if (segmentation.pixelCount (i) == 0) {
			toggleSelection (i);
		    }
		}
	    }
	}
	int countSelected () {
	    int count = 0;
	    for (int i = 0; i < selected.length(); i++) {
		if (selected (i)) {
		    count++;
		}
	    }
	    return count;
	}
	int[] getSelectedIDs () {
	    int n = countSelected();
	    int[] sel = new int [n];
	    int c = 0;
	    for (int i = 0; i < selected.length(); i++) {
		if (selected (i)) {
		    sel[c++] = i;
		}
	    }
	    return sel;
	}
    }
    
    static List undoActions = new ArrayList ();
    static List redoActions = new ArrayList ();

    static public void undo () {
	if (undoActions.size() > 0) {
	    EditAction editAction = (EditAction) undoActions.remove (undoActions.size() - 1);
	    redoActions.add (editAction);
	    editAction.undo ();
	    mainWindow.updateImageDisplay ();
	    mainWindow.splitterWindow.setSegments (editAction.leftSeg, editAction.rightSeg);
	    selection.selectAll (false);
	    if (editAction.leftSeg != -1) {
		selection.toggleSelection (editAction.leftSeg);
	    }
	    if (editAction.rightSeg != -1) {
		selection.toggleSelection (editAction.rightSeg);
	    }
	    mainWindow.splitterWindow.leftImagePanel.remaskImage (editAction.getClip());
	    mainWindow.splitterWindow.rightImagePanel.remaskImage (editAction.getClip());
	    mainWindow.splitterWindow.updateImageDisplay ();
	}
    }

    static public void redo () {
	if (redoActions.size() > 0) {
	    EditAction editAction = (EditAction) redoActions.remove (redoActions.size() - 1);
	    undoActions.add (editAction);
	    editAction.redo ();
	    mainWindow.updateImageDisplay ();
	    mainWindow.splitterWindow.setSegments (editAction.leftSeg, editAction.rightSeg);
	    selection.selectAll (false);
	    if (editAction.leftSeg != -1) {
		selection.toggleSelection (editAction.leftSeg);
	    }
	    if (editAction.rightSeg != -1) {
		selection.toggleSelection (editAction.rightSeg);
	    }
	    mainWindow.splitterWindow.leftImagePanel.remaskImage (editAction.getClip());
	    mainWindow.splitterWindow.rightImagePanel.remaskImage (editAction.getClip());
	    mainWindow.splitterWindow.updateImageDisplay ();
	}
    }

    static public void doEdit (EditAction editAction) {
	if (!editAction.empty()) {
	    editAction.redo ();
	    redoActions = new ArrayList ();
	    undoActions.add (editAction);
	    Rectangle clip = editAction.getClip();
	    mainWindow.splitterWindow.leftImagePanel.remaskImage (clip);
	    mainWindow.splitterWindow.rightImagePanel.remaskImage (clip);
	    mainWindow.splitterWindow.leftImagePanel.repaint ();
	    mainWindow.splitterWindow.rightImagePanel.repaint ();
	    mainWindow.updateUserInterface ();
	} else {
	    mainWindow.splitterWindow.repaint ();
	}
    }

    static class EditAction {
	List scanLines;
	Point tl, br;
	int leftSeg, rightSeg;

	class ScanLine {
	    int fromSeg, toSeg;
	    int row, colStart, colEnd;
	    public ScanLine (int fromSeg, int toSeg, int row, int colStart, int colEnd) {
		this.fromSeg = fromSeg;
		this.toSeg = toSeg;
		this.row = row;
		this.colStart = colStart;
		this.colEnd = colEnd;
	    }
	}

	public EditAction (int leftSeg, int rightSeg) {
	    this.leftSeg = leftSeg;
	    this.rightSeg = rightSeg;
	    scanLines = new ArrayList ();
	    tl = new Point (imWidth, imHeight);
	    br = new Point (0, 0);
	}

	public void addScanLine (int fromSeg, int toSeg, int row, int colStart, int colEnd) {
	    scanLines.add (new ScanLine (fromSeg, toSeg, row, colStart, colEnd));
	    tl.x = Math.min (tl.x, colStart);
	    tl.y = Math.min (tl.y, row);
	    br.x = Math.max (br.x, colEnd);
	    br.y = Math.max (br.y, row);
	}

	public Rectangle getClip () {
	    return new Rectangle (tl.x, tl.y, br.x - tl.x + 1, br.y - tl.y + 1);
	}

	public void redo () {
	    for (int i = 0; i < scanLines.size(); i++) {
		ScanLine scanLine = (ScanLine) scanLines.get (i);
		int y = scanLine.row;
		for (int x = scanLine.colStart; x <= scanLine.colEnd; x++) {
		    segmentation.set (x, y, scanLine.toSeg);
		}
	    }
	    unsavedChanges = true;	
	}

	public void undo () {
	    for (int i = 0; i < scanLines.size(); i++) {
		ScanLine scanLine = (ScanLine) scanLines.get (i);
		int y = scanLine.row;
		for (int x = scanLine.colStart; x <= scanLine.colEnd; x++) {
		    segmentation.set (x, y, scanLine.fromSeg);
		}
	    }
	    unsavedChanges = true;	
	}

	public boolean empty () {
	    return (scanLines.size() == 0);
	}
    }
}
