
// Author:  David Martin  <dmartin@eecs.berkeley.edu>

// Adapted from code in the Mailpuccino Free Java Email Client.
// Original copyright follows.

/* 
 * Mailpuccino Free Java Email Client
 * http://members.xoom.com/konget/mailpuccino/dev/
 * Copyright (C) 2000 Kong Eu Tak

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 */

import java.util.*;
import java.io.*;
import java.net.*;

// A class representing an SMTP server and a connection to it.
public class SmtpServer
{
    public static class Exception extends IOException {
	public Exception (String msg) { super (msg); }
    }
    public static class ErrorResponseException extends SmtpServer.Exception {
	public ErrorResponseException (String msg) { super (msg); }
    }
    public static class ConnectFailedException extends SmtpServer.Exception {
	public ConnectFailedException (String msg) { super (msg); }
    }
    public static class UnconnectedException extends SmtpServer.Exception {
	public UnconnectedException (String msg) { super (msg); }
    }

    final static int defaultPort = 25;

    String serverName;
    int port;

    Socket server = null;
    BufferedReader smtpReader = null;
    BufferedWriter smtpWriter = null;
	
    String lastReply;
    int lastReplyCode = 0;

    public SmtpServer (String serverName) {
	this (serverName, defaultPort);
    }

    public SmtpServer (String serverName, int port) {
	this.serverName = serverName;
	this.port = port;
    }

    public String getServerName () {
	return serverName; 
    }

    public int getPort () {
	return port;
    }

    public void connect () throws IOException {
	server = new Socket (getServerName(), getPort());
	smtpReader = new BufferedReader(
	    new InputStreamReader (server.getInputStream()));
	smtpWriter = new BufferedWriter(
	    new OutputStreamWriter (server.getOutputStream()));
	if (getResponse() != 220) {
	    throw new ConnectFailedException (lastReply);
	}
    }

    public void sendEmail (String sender, String recipient,
			   String subject, String message) throws IOException {
	sendCommand (250, "RSET");
	sendCommand (250, "MAIL FROM:" + sender);
	sendCommand (250, "RCPT TO:" + recipient);
	sendCommand (354, "DATA");
	sendMessage (recipient, subject, message);
    }

    public void sendEmail (String sender, String recipient,
			   String subject, Reader messageReader) throws IOException {
	sendCommand (250, "RSET");
	sendCommand (250, "MAIL FROM:" + sender);
	sendCommand (250, "RCPT TO:" + recipient);
	sendCommand (354, "DATA");
	sendMessage (recipient, subject, messageReader);
    }

    public void disconnect () {
	if (server == null) {
	    return;
	}
	try {
	    sendCommand (221, "QUIT");
	    server.close();
	    server = null;
	} catch (Throwable e){
	}
    }

    void sendCommand (int expected_reply, 
		      String CommandString) throws IOException {
	if (server == null) {
	    throw new UnconnectedException ("Not Connected to SMTP server.");
	}
	smtpWriter.write (CommandString);
	smtpWriter.write ("\r\n");
	smtpWriter.flush ();
	int reply = getResponse ();
	if (reply != expected_reply) {
	    throw new ErrorResponseException (lastReply); 
	}
    }

    void sendMessage (String recipient, String subject, 
		      String message) throws IOException {
	if (server == null) {
	    throw new UnconnectedException("Not Connected to SMTP server !");
	}
	smtpWriter.write ("To: " + recipient + "\r\n");
	smtpWriter.write ("Subject: " + subject + "\r\n");
	smtpWriter.write (message);
	smtpWriter.write ("\r\n");
	smtpWriter.flush ();
	sendCommand (250, ".");
    }

    void sendMessage (String recipient, String subject, 
		      Reader messageReader) throws IOException {
	if (server == null) {
	    throw new UnconnectedException("Not Connected to SMTP server !");
	}
	smtpWriter.write ("To: " + recipient + "\r\n");
	smtpWriter.write ("Subject: " + subject + "\r\n");
	while (true) {
	    int c = messageReader.read ();
	    if (c == -1) {
		break;
	    }
	    smtpWriter.write (c);
	}
	smtpWriter.write ("\r\n");
	smtpWriter.flush ();
	sendCommand (250, ".");
    }

    int getResponse() throws IOException {
	do {
	    lastReply = smtpReader.readLine();
	} while (lastReply.charAt(3) == '-');
	try {
	    lastReplyCode = Integer.parseInt (lastReply.substring (0, 3));
	} catch (IndexOutOfBoundsException e) {
	    throw new ErrorResponseException (lastReply); 
	} catch (NumberFormatException e) {
	    throw new ErrorResponseException (lastReply); 
	}
	return lastReplyCode;
    }
}

