
// Author:  David Martin  <dmartin@eecs.berkeley.edu>

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import javax.swing.*;
import javax.swing.border.*;
import java.util.*;
import java.awt.geom.*;
import java.util.List;

// A window that allows the user to bipartition an image.
class SplitterWindow
    extends JFrame
    implements ActionListener, AdjustmentListener, KeyListener
{
    static final int smallScroll = 20;
    static final int bigScroll = 100;

    static final int NO_SEGMENT = -1;
    int leftSeg = NO_SEGMENT;
    int rightSeg = NO_SEGMENT;

    MainWindow mainWindow;
    JMenuItem fileCloseItem, helpHelpItem;
    JMenuItem editMenuCancelSelection;
    JMenuItem editMenuUndoClick, editMenuRedoClick;
    JMenuItem editMenuUndo, editMenuRedo;
    JCheckBoxMenuItem viewMenuFitWindow;
    JMenuItem viewMenuRefresh;
    JMenuItem viewMenuMag1x, viewMenuMag2x, viewMenuMag3x, viewMenuMag4x;
    JTextField magTextField;
    JCheckBox fitWindowCheckBox, vertSplitCheckBox;
    JViewport leftViewport, rightViewport;
    JPanel leftViewportPanel, rightViewportPanel;
    ImagePanel leftImagePanel, rightImagePanel;
    JScrollBar vsb, hsb;
    JPanel imagePanel;

    // Magnification.
    static boolean fitWindow = true;
    static double magnification = 1.0;

    // Vertical or horizontal split window layout.
    boolean vertSplit = true;

    // The size of the border placed around the image in the ImagePanel.
    final static int border = 15;

    public SplitterWindow (MainWindow mainWindow, String title) {
	super (title);
	this.mainWindow = mainWindow;

	JPanel tmpPanel;

	JMenu fileMenu = new JMenu ("File");
	fileCloseItem = new JMenuItem ("Close");
	fileMenu.add (fileCloseItem);
	fileCloseItem.addActionListener (this);
	fileCloseItem.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_W, ActionEvent.CTRL_MASK));

	JMenu editMenu = new JMenu ("Edit");
	editMenuCancelSelection = new JMenuItem ("Cancel Selection");
	editMenuUndoClick = new JMenuItem ("Undo Click");
	editMenuRedoClick = new JMenuItem ("Redo Click");
	editMenuUndo = new JMenuItem ("Undo");
	editMenuRedo = new JMenuItem ("Redo");
	editMenuCancelSelection.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_C, ActionEvent.CTRL_MASK));
	editMenuUndoClick.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_BACK_SPACE, 0));
	editMenuRedoClick.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_ENTER, 0));
	editMenuUndo.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_Z, ActionEvent.CTRL_MASK));
	editMenuRedo.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_R, ActionEvent.CTRL_MASK));
	editMenu.add (editMenuCancelSelection);
	editMenu.addSeparator ();
	editMenu.add (editMenuUndoClick);
	editMenu.add (editMenuRedoClick);
	editMenu.addSeparator ();
	editMenu.add (editMenuUndo);
	editMenu.add (editMenuRedo);
	editMenuCancelSelection.addActionListener (this);
	editMenuUndoClick.addActionListener (this);
	editMenuRedoClick.addActionListener (this);
	editMenuUndo.addActionListener (this);
	editMenuRedo.addActionListener (this);

	JMenu helpMenu = new JMenu ("Help");
	helpHelpItem = new JMenuItem ("Help");
	helpMenu.add (helpHelpItem);
	helpHelpItem.addActionListener (this);
	fileCloseItem.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_H, ActionEvent.CTRL_MASK));

	JMenu viewMenu = new JMenu ("View");
	viewMenuRefresh = new JMenuItem ("Refresh Window");
	viewMenuFitWindow = new JCheckBoxMenuItem ("Fit Window");
	viewMenuMag1x = new JMenuItem ("Magnify 1x");
	viewMenuMag2x = new JMenuItem ("Magnify 2x");
	viewMenuMag3x = new JMenuItem ("Magnify 3x");
	viewMenuMag4x = new JMenuItem ("Magnify 4x");
	viewMenuFitWindow.setSelected (fitWindow);
	viewMenuRefresh.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_L, ActionEvent.CTRL_MASK));
	viewMenuFitWindow.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_F, ActionEvent.CTRL_MASK));
	viewMenuMag1x.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_1, ActionEvent.CTRL_MASK));
	viewMenuMag2x.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_2, ActionEvent.CTRL_MASK));
	viewMenuMag3x.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_3, ActionEvent.CTRL_MASK));
	viewMenuMag4x.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_4, ActionEvent.CTRL_MASK));
	viewMenu.add (viewMenuRefresh);
	viewMenu.addSeparator ();
	viewMenu.add (viewMenuMag1x);
	viewMenu.add (viewMenuMag2x);
	viewMenu.add (viewMenuMag3x);
	viewMenu.add (viewMenuMag4x);
	viewMenu.add (viewMenuFitWindow);
	viewMenuMag1x.addActionListener (this);
	viewMenuMag2x.addActionListener (this);
	viewMenuMag3x.addActionListener (this);
	viewMenuMag4x.addActionListener (this);
	viewMenuFitWindow.addActionListener (this);
	viewMenuRefresh.addActionListener (this);

	JMenuBar mbar = new JMenuBar ();
	mbar.add (fileMenu);
	mbar.add (editMenu);
	mbar.add (viewMenu);
	mbar.add (Box.createHorizontalGlue ());
	mbar.add (helpMenu);
	setJMenuBar (mbar);

	JLabel magLabel = new JLabel ("Magnification: ");
	magLabel.setForeground (Color.black);
	magTextField = new JTextField (5);
	magTextField.addActionListener (this);
	fitWindowCheckBox = new JCheckBox ("Fit Window", fitWindow);
	fitWindowCheckBox.addActionListener (this);
	vertSplitCheckBox = new JCheckBox ("Vertical Split", vertSplit);
	vertSplitCheckBox.addActionListener (this);

	JPanel magTextBox = new JPanel ();
	magTextBox.setLayout (new FlowLayout (FlowLayout.LEFT));
	magTextBox.add (magTextField);

	Box headBox = Box.createHorizontalBox ();
	headBox.add (Box.createHorizontalStrut (5));
	headBox.add (magLabel);
	headBox.add (Box.createHorizontalStrut (5));
	headBox.add (magTextBox);
	headBox.add (Box.createHorizontalGlue ());
	headBox.add (Box.createHorizontalStrut (5));
	headBox.add (fitWindowCheckBox);
	headBox.add (Box.createHorizontalStrut (5));
	headBox.add (vertSplitCheckBox);
	headBox.add (Box.createHorizontalStrut (5));
	JPanel headPanel = new JPanel ();
	headPanel.setLayout (new BorderLayout ());
	headPanel.add (headBox, BorderLayout.NORTH);
	headPanel.setBorder (new BevelBorder (BevelBorder.RAISED));

	leftImagePanel = new ImagePanel ();
	rightImagePanel = new ImagePanel ();
	leftImagePanel.setPeer (rightImagePanel);
	rightImagePanel.setPeer (leftImagePanel);

	this.addKeyListener (this);

	vsb = new JScrollBar (Adjustable.VERTICAL);
	hsb = new JScrollBar (Adjustable.HORIZONTAL);
	vsb.addAdjustmentListener (this);
	hsb.addAdjustmentListener (this);

	leftViewport = new JViewport ();
	leftViewport.setView (leftImagePanel);
	rightViewport = new JViewport ();
	rightViewport.setView (rightImagePanel);

	leftViewportPanel = new JPanel ();
	leftViewportPanel.setLayout (new BorderLayout ());
	leftViewportPanel.add (leftViewport, BorderLayout.CENTER);
	leftViewportPanel.setBorder (new EtchedBorder (EtchedBorder.RAISED));
	rightViewportPanel = new JPanel ();
	rightViewportPanel.setLayout (new BorderLayout ());
	rightViewportPanel.add (rightViewport, BorderLayout.CENTER);
	rightViewportPanel.setBorder (new EtchedBorder (EtchedBorder.RAISED));

	Box hsbBox = Box.createHorizontalBox ();
	hsbBox.add (hsb);
	hsbBox.add (new ScrollCorner (hsb, vsb));

	imagePanel = new JPanel ();
	setImagePanelLayout ();

	JPanel midPanel = new JPanel ();
	midPanel.setLayout (new BorderLayout ());
	midPanel.add (hsbBox, BorderLayout.SOUTH);
	midPanel.add (vsb, BorderLayout.EAST);
	midPanel.add (imagePanel, BorderLayout.CENTER);

	JPanel mainPanel = new JPanel ();
	mainPanel.setLayout (new BorderLayout ());
	mainPanel.add (headPanel, BorderLayout.NORTH);
	mainPanel.add (midPanel, BorderLayout.CENTER);

	setContentPane (mainPanel);

	setDefaultCloseOperation (JFrame.DO_NOTHING_ON_CLOSE);
	addWindowListener (new WindowAdapter () {
	    public void windowClosing (WindowEvent e) {
		hide ();
	    };
	});

	pack ();	// Must call pack before setSize.
	setSize (800,800);
	setLocation (500, 0);

	leftViewport.addComponentListener (
	    new ComponentAdapter () {
		    public void componentResized (ComponentEvent e) {
			updateImageDisplay ();
		    }
		});
    }

    void setImagePanelLayout () {
	imagePanel.remove (leftViewportPanel);
	imagePanel.remove (rightViewportPanel);
	if (vertSplit) {
	    imagePanel.setLayout (new GridLayout (1, 2));
	} else {
	    imagePanel.setLayout (new GridLayout (2, 1));
	}
	imagePanel.add (leftViewportPanel);
	imagePanel.add (rightViewportPanel);
    }

    // Call this method when we load a new image.
    // Preconditions: An image is loaded.
    // Actions:  Set up data structures that depend on the image size.
    // Postconditions: Can call setSegments().
    public void newImage () {
	Util.assert (SegmentApp.imageLoaded());
	if ((SegmentApp.imWidth > SegmentApp.imHeight) == vertSplit) {
	    vertSplit = !vertSplit; 
	    setImagePanelLayout(); 
	}
	scratchImage = null;
	if (scratchGraphics != null) { scratchGraphics.dispose (); }
	scratchGraphics = null;
	leftImagePanel.newImage ();
	rightImagePanel.newImage ();
	clearSegments ();
    }

    // Call this method to change the two segments between which we
    // want to exchange pixels.
    // Preconditions: 
    //	Image must be loaded.
    //  newImage() must have been called.
    // Actions: Update the UI to display the two segments.
    // Postconditions: User can move pixels between segments.
    public void setSegments (int a, int b) {
	Util.assert (SegmentApp.imageLoaded());
	leftSeg = a;
	rightSeg = b;
	leftImagePanel.setSegment (a);
	rightImagePanel.setSegment (b);
    }
    public void clearSegments () {
	leftSeg = NO_SEGMENT;
	rightSeg = NO_SEGMENT;
	leftImagePanel.clearSegment ();
	rightImagePanel.clearSegment ();
    }

    public void updateMenus () {
	editMenuUndo.setEnabled (SegmentApp.undoActions.size() > 0 
				 && !leftImagePanel.selectionActive()
				 && !rightImagePanel.selectionActive());
	editMenuRedo.setEnabled (SegmentApp.redoActions.size() > 0 
				 && !leftImagePanel.selectionActive()
				 && !rightImagePanel.selectionActive());
	editMenuCancelSelection.setEnabled (
	    leftImagePanel.selectionActive()
	    || rightImagePanel.selectionActive());
	editMenuUndoClick.setEnabled (
	    leftImagePanel.selectionActive()
	    || rightImagePanel.selectionActive());
	editMenuRedoClick.setEnabled (
	    leftImagePanel.selectionActive()
	    || rightImagePanel.selectionActive());
	viewMenuFitWindow.setSelected (fitWindow);
	fitWindowCheckBox.setSelected (fitWindow);
	vertSplitCheckBox.setSelected (vertSplit);
    }

    public void updateScrollbars () {
	Dimension viewDim = leftViewport.getViewSize ();
	Dimension portDim = leftViewport.getExtentSize ();
	Point portLoc = leftViewport.getViewPosition ();

	vsb.setMinimum (0);
	vsb.setMaximum (viewDim.height);
	vsb.setValue (portLoc.y);
	vsb.setVisibleAmount (portDim.height);
	vsb.setUnitIncrement (smallScroll);
	vsb.setBlockIncrement (bigScroll);

	hsb.setMinimum (0);
	hsb.setMaximum (viewDim.width);
	hsb.setValue (portLoc.x);
	hsb.setVisibleAmount (portDim.width);
	hsb.setUnitIncrement (smallScroll);
	hsb.setBlockIncrement (bigScroll);
    }

    // Make sure UI is in sync with data structures.
    public void updateUserInterface () {
	magTextField.setText (String.valueOf (magnification));
	updateMenus ();
	validate ();  
	repaint ();
    }
    public void updateImageDisplay () {
	if (SegmentApp.imageLoaded() && fitWindow) {
	    setMagnification (magFromViewportSize ());
	}
	leftImagePanel.invalidate ();
	rightImagePanel.invalidate ();
	validate ();
	updateScrollbars ();
	updateUserInterface ();
    }
    
    public void adjustmentValueChanged (AdjustmentEvent e) {
	Adjustable object = e.getAdjustable ();
	int x = hsb.getValue ();
	int width = hsb.getVisibleAmount ();
	int y = vsb.getValue ();
	int height = vsb.getVisibleAmount ();
	leftViewport.setViewPosition (new Point (x, y));
	leftViewport.setExtentSize (new Dimension (width, height));
	rightViewport.setViewPosition (new Point (x, y));
	rightViewport.setExtentSize (new Dimension (width, height));
    }

    // Arrow keys scroll the window.
    public void keyPressed (KeyEvent e) {
	int code = e.getKeyCode ();
	int step = (e.isShiftDown() || e.isControlDown() || e.isAltDown()) 
		   ? bigScroll : smallScroll;
	Point pos = leftViewport.getViewPosition ();
	Dimension view = leftViewport.getViewSize ();
	Dimension extent = leftViewport.getExtentSize();
	boolean changed = false;
	Point delta = new Point (0, 0);
	Point p = new Point (pos);
	switch (code) {
	    case KeyEvent.VK_UP: 
		p.y = Math.max (0, p.y - step);
		changed = true;
		break;
	    case KeyEvent.VK_DOWN: 
		p.y = Math.min (view.height - extent.height, p.y + step);
		changed = true;
		break;
	    case KeyEvent.VK_LEFT:
		p.x = Math.max (0, p.x - step);
		changed = true;
		break;
	    case KeyEvent.VK_RIGHT:
		p.x = Math.min (view.width - extent.width, p.x + step);
		changed = true;
		break;
	}
	if (changed) {
	    leftViewport.setViewPosition (p);
	    rightViewport.setViewPosition (p);
	    updateScrollbars ();
	    delta.x = p.x - pos.x;
	    delta.y = p.y - pos.y;
	    leftImagePanel.shiftRubberBand (delta);
	    rightImagePanel.shiftRubberBand (delta);
	}
    }
    public void keyReleased (KeyEvent e) {
    }
    public void keyTyped (KeyEvent e) {
    }
    
    public void actionPerformed (ActionEvent e) {
	Object item = e.getSource ();
	if (item == fileCloseItem) {
	    hide ();
	} else if (item == editMenuCancelSelection) {
	    leftImagePanel.cancelSelection ();
	    rightImagePanel.cancelSelection ();
	} else if (item == editMenuUndoClick) {
	    leftImagePanel.undoClick ();
	    rightImagePanel.undoClick ();
	} else if (item == editMenuRedoClick) {
	    leftImagePanel.redoClick ();
	    rightImagePanel.redoClick ();
	} else if (item == editMenuUndo) {
	    SegmentApp.undo ();
	} else if (item == editMenuRedo) {
	    SegmentApp.redo ();
	} else if (item == helpHelpItem) {
	    help ();
	} else if (item == viewMenuRefresh) {
	    // UI updated below.
	} else if (item == viewMenuFitWindow) {
	    fitWindow = viewMenuFitWindow.isSelected ();
	    if (fitWindow) {
		updateImageDisplay ();
	    }
	} else if (item == viewMenuMag1x) {
	    setMagnification (1);
	    fitWindow = false;
	    updateImageDisplay ();
	} else if (item == viewMenuMag2x) {
	    setMagnification (2);
	    fitWindow = false;
	    updateImageDisplay ();
	} else if (item == viewMenuMag3x) {
	    setMagnification (3);
	    fitWindow = false;
	    updateImageDisplay ();
	} else if (item == viewMenuMag4x) {
	    setMagnification (4);
	    fitWindow = false;
	    updateImageDisplay ();
	} else if (item == fitWindowCheckBox) {
	    fitWindow = fitWindowCheckBox.isSelected ();
	    if (fitWindow) {
		updateImageDisplay ();
	    }
	} else if (item == vertSplitCheckBox) {
	    vertSplit = vertSplitCheckBox.isSelected ();
	    setImagePanelLayout ();
	    validate ();
	} else if (item == magTextField) {
	    // User entered a new magnification in the text area.
	    String text = magTextField.getText();
	    double newMag;
	    try {
		newMag = Double.parseDouble (text);
	    } catch (NumberFormatException exn) {
		newMag = magnification;
	    }
	    setMagnification (newMag);
	    fitWindow = false;
	    updateImageDisplay ();
	}
	updateUserInterface ();
    }

    // Always call this method to change the magnification.
    void setMagnification (double mag) {
	// Make sure magnification is in the valid range.
	mag = Math.max (mag, SegmentApp.minMagnification);
	mag = Math.min (mag, SegmentApp.maxMagnification);
	// Make sure magnification has <= 5 significant digits
	// for display purposes.
	mag = (double) ((int) (mag * 10000.)) / 10000.;
	magnification = mag;
    }

    // Return the magnification that causes the image to fit perfectly
    // in the viewport.
    double magFromViewportSize () {
	Util.assert (SegmentApp.imageLoaded());
	double mag;
	Dimension dim = leftViewport.getExtentSize();
	if (dim.width < SegmentApp.imWidth * SegmentApp.minMagnification + 2 * border 
	    || dim.height < SegmentApp.imHeight * SegmentApp.minMagnification + 2 * border) {
	    mag = SegmentApp.minMagnification;
	} else {
	    double imageAspectRatio = (double) SegmentApp.imWidth / SegmentApp.imHeight;
	    double panelAspectRatio = (double) (dim.width - 2 * border) 
				      / (dim.height - 2 * border);
	    if (imageAspectRatio > panelAspectRatio) {
		mag = (double) (dim.width - 2 * border) / SegmentApp.imWidth;
	    } else {
		mag = (double) (dim.height - 2 * border) / SegmentApp.imHeight;
	    }
	}
	return mag;
    }

    void help () {
	final String message = 
	    "OVERVIEW " +
	    "\n" +
	    "The purpose of this window is to divide the pixels of some image " +
	    "segment into two groups.  Selecting a set of pixels in one group " +
	    "causes those pixels to be transferred to the other group. " +
	    "\n" +
	    "You select a set of pixels by tracing a boundary contour with the " +
	    "mouse. A mouse press with the left mouse button starts the " +
	    "contour. Dragging the mouse while holding down the left mouse button " +
	    "will extend the contour in a free-form manner.  If you release the " +
	    "left mouse button while creating the contour, the contour will be " +
	    "extended with a straight line segment to the point at which you next " +
	    "press the left mouse button. " +
	    "\n" +
	    "Either a middle button click, a right button click, or a click of the " +
	    "left button while holding down the Control key will close the contour " +
	    "and move the enclosed pixels to the other group.  You can cancel an " +
	    "ongoing selection with the \"Cancel Selection\" command from the Edit " +
	    "menu. " +
	    "\n" +
	    "The Undo (Ctrl-Z) and Redo (Ctrl-R) commands apply to actions that " +
	    "move pixels between the panels in the splitter window.  The Undo Click " +
	    "(Backspace) and Redo Click (Enter) commands provide fine-grained " +
	    "undo/redo while you are tracing a contour. " +
	    "\n" +
	    "TIPS " +
	    "\n" +
	    "Make sure your mouse is clean. " +
	    "\n" +
	    "Use the zoom!  One particularly effective mode of operation is to " +
	    "switch between fit window (Ctrl-F) to get your bearings, and 3x or 4x " +
	    "magnification (Ctrl-3 or Ctrl-4) to accurately trace contours.  Note " +
	    "that you can also zoom the window while tracing a contour. " +
	    "\n" +
	    "Drawing contours free-hand is rarely useful.  It's often faster and " +
	    "more accurate to trace boundaries with many small clicks. " +
	    "\n" +
	    "You can scroll and zoom the window while in the midst of tracing a " +
	    "contour. " +
	    "\n";
	JOptionPane.showMessageDialog (
	    this, Util.lineBreak (message, SegmentApp.helpWidth),
	    "Help", JOptionPane.INFORMATION_MESSAGE);
    }

    // Share these across the two instances of ImagePanel, since the
    // image takes up a lot of memory for large images.
    BufferedImage scratchImage;
    Graphics2D scratchGraphics;

    // The segment ID assigned to an ImagePanel object may or may not be valid,
    // so this class should make no assumptions about the validity of the ID.
    class ImagePanel 
	extends JPanel 
	implements MouseListener, MouseMotionListener
    {
	int seg = NO_SEGMENT;
	ImagePanel peer;
	BufferedImage maskedImage;
	Point2D.Float firstPoint, lastPoint, rubberPoint;
	List selectionPoints, undonePoints;

	public ImagePanel () {
	    addMouseListener (this);
	    addMouseMotionListener (this);
	    //setCursor (new Cursor (Cursor.CROSSHAIR_CURSOR));
	}

	public void setPeer (ImagePanel peer) {
	    this.peer = peer;
	}

	public void newImage () {
	    Util.assert (SegmentApp.imageLoaded());
	    maskedImage = new BufferedImage (
		SegmentApp.imWidth, SegmentApp.imHeight, 
		BufferedImage.TYPE_INT_ARGB);
	    if (scratchImage == null) {
		scratchImage = new BufferedImage (
		    SegmentApp.imWidth, SegmentApp.imHeight, 
		    BufferedImage.TYPE_INT_ARGB);
		scratchGraphics = scratchImage.createGraphics ();
	    }
	    Util.assert (scratchImage != null);
	    Util.assert (scratchGraphics != null);
	}

	public void clearSegment () {
	    this.seg = NO_SEGMENT;
	    if (SegmentApp.imageLoaded()) {
		remaskImage ();
	    }
	}

	public void setSegment (int seg) {
	    Util.assert (SegmentApp.imageLoaded());
	    this.seg = seg;
	    remaskImage ();
	}

	public Dimension getPreferredSize () {
	    if (SegmentApp.imageLoaded()) {
		return new Dimension (
		    (int) Math.round ((double)SegmentApp.imWidth * magnification + 2 * border), 
		    (int) Math.round ((double)SegmentApp.imHeight * magnification + 2 * border));
	    } else {
		return super.getPreferredSize ();
	    }
	}

	// Window coordinates are integral, and correspond to screen pixels.
	// Image coordinates are floats.  They are corrected for magnification.
	Point2D.Float windowToImage (Point p) {
	    Dimension d = getSize ();
	    float x = (float) p.x;
	    float y = (float) p.y;
	    // Jam coordinates to window size.
	    x = Math.max (0, Math.min (x, d.width - 1));
	    y = Math.max (0, Math.min (y, d.height - 1));
	    // Convert from window to image coordinates.
	    x = (float) ((x - border) / magnification);
	    y = (float) ((y - border) / magnification);
	    return new Point2D.Float (x, y);
	}
	Point imageToWindow (Point2D.Float p) {
	    // Convert from window to image coordinates.
	    return new Point (
		(int) Math.round (border + p.x * magnification),
		(int) Math.round (border + p.y * magnification));
	}
	void addSelectionPoint (Point2D.Float p) {
	    Util.assert (SegmentApp.imageLoaded());
	    selectionPoints.add (p);
	    undonePoints = new ArrayList ();
	}
	void startSelection (Point2D.Float p) {
	    Util.assert (SegmentApp.imageLoaded());
	    // If peer has a selection going, then don't start one here!
	    if (peer.selectionActive()) {
		return;
	    }
	    selectionPoints = new ArrayList ();
	    addSelectionPoint (p);
	    Graphics g = getGraphics ();
	    g.setColor (Color.green);
	    Point a = imageToWindow (p);
	    g.drawLine (a.x, a.y, a.x, a.y);
	    firstPoint = lastPoint = p;
	    rubberPoint = null;
	    g.dispose ();
	}
	// Bresenham's line drawing algorithm, taken from:
	// http://graphics.lcs.mit.edu/~mcmillan/comp136/Lecture6/Lines.html
	// The line-drawing routine in the AWT is actually buggy.  When I
	// draw lines using graphics contexts from different sources (e.g.
	// passed in to the repaint() routine, or gotten with getGraphics()),
	// I get different lines, which leaves artifacts when redrawing
	// rubber-bands.
	void myDrawPixel (Graphics g, int x, int y) {
	    g.drawLine (x, y, x, y);
	}
	void myDrawLine (Graphics g, Point a, Point b) {
	    int x0 = a.x;
	    int y0 = a.y;
	    int x1 = b.x;
	    int y1 = b.y;

	    int dy = y1 - y0;
	    int dx = x1 - x0;
	    int stepx, stepy;
	    
	    if (dy < 0) { dy = -dy;  stepy = -1; } else { stepy = 1; }
	    if (dx < 0) { dx = -dx;  stepx = -1; } else { stepx = 1; }
	    dy <<= 1;                             // dy is now 2*dy
	    dx <<= 1;                             // dx is now 2*dx
	  
	    myDrawPixel (g, x0, y0);
	    if (dx > dy) {
		int fraction = dy - (dx >> 1);    // same as 2*dy - dx
		while (x0 != x1) {
		    if (fraction >= 0) {
			y0 += stepy;
			fraction -= dx;           // same as fraction -= 2*dx
		    }
		    x0 += stepx;
		    fraction += dy;               // same as fraction -= 2*dy
		    myDrawPixel (g, x0, y0);
		}
	    } else {
		int fraction = dx - (dy >> 1);
		while (y0 != y1) {
		    if (fraction >= 0) {
			x0 += stepx;
			fraction -= dy;
		    }
		    y0 += stepy;
		    fraction += dx;
		    myDrawPixel (g, x0, y0);
		}
	    }
	}
	void shiftRubberBand (Point delta) {
	    Point p = imageToWindow (rubberPoint);
	    p.x += delta.x;
	    p.y += delta.y;
	    rubberBandSelection (windowToImage (p));
	}
	void drawRubberBand (Graphics g) {
	    if (rubberPoint == null) { return; }
	    Util.assert (lastPoint != null);
	    g.setColor (Color.red);
	    g.setXORMode (Color.green);
	    Point a = imageToWindow (lastPoint);
	    Point b = imageToWindow (rubberPoint);
	    myDrawLine (g, a, b);
	}
	// Erase the previous rubber-banded line.
	// Set rubberPoint to null.
	void unRubberBand () {
	    Util.assert (SegmentApp.imageLoaded());
	    if (rubberPoint == null) { return; }
	    Graphics g = getGraphics ();
	    drawRubberBand (g);
	    rubberPoint = null;
	    g.dispose ();
	}
	// Install new rubber-band point.
	// Erase previous rubber-band line, and draw new one.
	void rubberBandSelection (Point2D.Float p) {
	    Util.assert (SegmentApp.imageLoaded());
	    Util.assert (p != null);
	    Graphics g = getGraphics ();
	    drawRubberBand (g);
	    rubberPoint = p;
	    drawRubberBand (g);
	    g.dispose ();
	}
	void extendSelection (Point2D.Float p) {
	    Util.assert (SegmentApp.imageLoaded());
	    unRubberBand ();
	    addSelectionPoint (p);
	    Graphics g = getGraphics ();
	    g.setColor (Color.green);
	    Point a = imageToWindow (lastPoint);
	    Point b = imageToWindow (p);
	    g.drawLine (a.x, a.y, b.x, b.y);
	    lastPoint = p;
	    g.dispose ();
	}
	void endSelection () {
	    Util.assert (SegmentApp.imageLoaded());
	    unRubberBand ();
	    Graphics g = getGraphics ();
	    g.setColor (Color.green);
	    Point a = imageToWindow (firstPoint);
	    Point b = imageToWindow (lastPoint);
	    g.drawLine (a.x, a.y, b.x, b.y);
	    g.dispose ();

	    excludePixels ();
	    nullifySelection ();
	}
	void cancelSelection () {
	    if (!selectionActive()) { return; }
	    Util.assert (SegmentApp.imageLoaded());
	    nullifySelection ();
	}
	void undoClick () {
	    if (!selectionActive()) { return; }
	    Util.assert (SegmentApp.imageLoaded());
	    int n = selectionPoints.size();
	    Util.assert (n > 0);
	    if (n == 1) { 
		cancelSelection (); 
		return;
	    }
	    undonePoints.add (selectionPoints.remove (n - 1));
	    lastPoint = (Point2D.Float) selectionPoints.get (n - 2);
	}
	void redoClick () {
	    if (!selectionActive()) { return; }
	    Util.assert (SegmentApp.imageLoaded());
	    int n = undonePoints.size();
	    if (n == 0) { return; }
	    lastPoint = (Point2D.Float) undonePoints.remove (n - 1);
	    selectionPoints.add (lastPoint);
	}
	void nullifySelection () {
	    Util.assert (SegmentApp.imageLoaded());
	    selectionPoints = null;
	    undonePoints = null;
	}
	boolean selectionActive () {
	    return selectionPoints != null;
	}

	public void mousePressed (MouseEvent e) {
	    if (!SegmentApp.imageLoaded()) {
		return;
	    }
	    if ((e.getModifiers() & e.BUTTON1_MASK) != 0) {
		if (selectionActive()) {
		    extendSelection (windowToImage (e.getPoint ()));
		    if ((e.getModifiers() & e.CTRL_MASK) != 0) {
			endSelection ();
			updateUserInterface ();
		    }
		} else {
		    startSelection (windowToImage (e.getPoint ()));
		    updateUserInterface ();
		}
	    } else if ((e.getModifiers() & e.BUTTON2_MASK) != 0) {
		if (selectionActive()) {
		    extendSelection (windowToImage (e.getPoint ()));
		    endSelection ();
		    updateUserInterface ();
		}
	    } else if ((e.getModifiers() & e.BUTTON3_MASK) != 0) {
		if (selectionActive()) {
		    extendSelection (windowToImage (e.getPoint ()));
		    endSelection ();
		    updateUserInterface ();
		}
	    }
	}
	public void mouseDragged (MouseEvent e) {
	    if (!SegmentApp.imageLoaded()) {
		return;
	    }
	    if ((e.getModifiers() & e.BUTTON1_MASK) != 0) {
		if (selectionActive()) {
		    extendSelection (windowToImage (e.getPoint ()));
		}
	    }
	}
	public void mouseClicked (MouseEvent e) {
	}
	public void mouseMoved (MouseEvent e) {
	    if (!SegmentApp.imageLoaded()) {
		return;
	    }
	    if (selectionActive()) {
		rubberBandSelection (windowToImage (e.getPoint()));
	    }
	}
	public void mouseReleased (MouseEvent e) {
	}
	public void mouseEntered (MouseEvent e) {
	    if (selectionActive()) {
		// Redraw rubber-band on mouse enter.
		rubberBandSelection (windowToImage (e.getPoint()));
	    }
	}
	public void mouseExited (MouseEvent e) {
	    if (selectionActive()) {
		// Erase rubber-band on mouse exit.
		unRubberBand ();
	    }
	}

	// Move all pixels in the interior of the polygon to 
	// our peer.
	// Construct an EditAction object for undo/redo.
	void excludePixels () {
	    Util.assert (SegmentApp.imageLoaded());
	    // Create a path object from the selection point list.
	    Util.assert (selectionPoints.size() > 0);
	    GeneralPath selection = new GeneralPath (GeneralPath.WIND_NON_ZERO);
	    Point2D.Float p = (Point2D.Float) selectionPoints.get (0);
	    selection.moveTo (p.x, p.y);
	    for (int i = 1; i < selectionPoints.size(); i++) {
		p = (Point2D.Float) selectionPoints.get (i);
		selection.lineTo (p.x, p.y);
	    }
	    selection.closePath ();
	    int blackRGB = Color.black.getRGB();
	    // Compute a bounding box for the polygon, and clip it
	    // to the image size.
	    Rectangle bounds = selection.getBounds();
	    int xmin = (int) Math.max (0, bounds.x - 1);
	    int xmax = (int) Math.min (bounds.x + bounds.width + 1, SegmentApp.imWidth);
	    int ymin = (int) Math.max (0, bounds.y - 1);
	    int ymax = (int) Math.min (bounds.y + bounds.height + 1, SegmentApp.imHeight);
	    // Fill the bounding box with white.
            scratchGraphics.setColor (Color.white);
	    scratchGraphics.fillRect (xmin, ymin, xmax - xmin, ymax - ymin);
	    // Draw the filled polygon in black.
	    scratchGraphics.setColor (Color.black);
	    scratchGraphics.fill (selection);
	    // Find all the black pixels.
	    // Scan-line convert transferred pixels for undo/redo.
	    SegmentApp.EditAction editAction = new SegmentApp.EditAction (leftSeg, rightSeg);
	    for (int y = ymin; y < ymax; y++) {
		int xstart = -1;
		int xend = -1;
		for (int x = xmin; x < xmax; x++) {
		    if (scratchImage.getRGB (x, y) == blackRGB
			&& SegmentApp.segmentation.get (x, y) == seg) {
			// Start new or extend current scan line.
			if (xstart == -1) {
			    xstart = xend = x;	// Starting a new scan line.
			} else {
			    Util.assert (x == xend + 1);
			    xend = x;		// Extending the scan line.
			}
		    } else {
			// End any ongoing scan line.
			if (xstart != -1) {
			    editAction.addScanLine (seg, peer.seg, y, xstart, xend);
			}
			xstart = xend = -1;
		    }
		}
		if (xstart != -1) {
		    editAction.addScanLine (seg, peer.seg, y, xstart, xend);
		}
	    }
	    SegmentApp.doEdit (editAction);
	}

	// Recompute the specified area of the masked image.
	public void remaskImage () {
	    Util.assert (SegmentApp.imageLoaded());
	    remaskImage (new Rectangle (SegmentApp.imWidth, SegmentApp.imHeight));
	}
	public void remaskImage (Rectangle clip) {
	    Util.assert (SegmentApp.imageLoaded());
	    int xmin = Math.max (0, clip.x);
	    int xmax = Math.min (clip.x + clip.width, SegmentApp.imWidth);
	    int ymin = Math.max (0, clip.y);
	    int ymax = Math.min (clip.y + clip.height, SegmentApp.imHeight);
	    for (int y = ymin; y < ymax; y++) {
		for (int x = xmin; x < xmax; x++) {
		    int rgb = SegmentApp.displayImage.getRGB (x, y);
		    if (SegmentApp.segmentation.get (x, y) != seg) {
			maskedImage.setRGB (x, y, RGBFilter.lighten (rgb));
		    } else {
			maskedImage.setRGB (x, y, rgb);
		    }
		}
	    }
	}

	public void paintSelection (Graphics g) {
	    if (selectionActive()) {
		g.setColor (Color.green);
		for (int i = 1; i < selectionPoints.size(); i++) {
		    Point2D.Float a = (Point2D.Float) selectionPoints.get (i - 1);
		    Point2D.Float b = (Point2D.Float) selectionPoints.get (i);
		    g.drawLine (
			(int) Math.round (border + a.x * magnification), 
			(int) Math.round (border + a.y * magnification), 
			(int) Math.round (border + b.x * magnification), 
			(int) Math.round (border + b.y * magnification)); 
		}
		drawRubberBand (g);
	    }
	}

	public void paint (Graphics g) {
	    super.paint (g);
	    if (SegmentApp.imageLoaded()) {
		int w = (int) Math.round ((double) SegmentApp.imWidth * magnification);
		int h = (int) Math.round ((double) SegmentApp.imHeight * magnification);
		boolean loaded = g.drawImage (
		    maskedImage, 
		    border, border,
		    w, h, null);
		Util.assert (loaded);
		paintSelection (g);
	    }
	}
    }
}
