
// Author:  David Martin  <dmartin@eecs.berkeley.edu>

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.util.*;
import java.io.*;
import java.net.*;
import gnu.regexp.*;
import java.util.List;

public class Util {
    // Assertions.
    // Assert failures are catastrophic.
    static void assert (boolean b) {
	assert (b, "");
    }
    static void assert (boolean b, String msg) {
	if (!b) { throw new InternalError (msg); }
    }

    // Preconditions: None.
    // Actions: Wait until the given image has completely loaded.
    // Return Value: true on any error; false on success.
    static boolean waitForImage (Image image) {
	MediaTracker tracker = new MediaTracker (new Container ());
	try {
	    tracker.addImage (image, 0);
	    tracker.waitForID (0);
	}
	catch (InterruptedException e) {
	    return true;
	}
	return tracker.isErrorID (0);
    }

    // Preconditions: None.
    // Actions: Load and return the named image.  
    static Image loadImage (File file) throws Exception {
	String imageFileName = file.getAbsolutePath();
	Image image = Toolkit.getDefaultToolkit().createImage (imageFileName);
	boolean error = waitForImage (image);
	if (error) { 
	    throw new Exception (
		"Could not load image file '" + file + "'."); 
	}
	return image;
    }
    static Image loadImage (URL url) throws Exception {
	Image image = Toolkit.getDefaultToolkit().createImage (url);
	boolean error = waitForImage (image);
	if (error) { 
	    throw new Exception (
		"Could not load image URL '" + url + "'."); 
	}
	return image;
    }

    // Break input string into an array of strings with at most width characters per
    // string.  Any newlines in the input are converted into blank output lines.
    // TODO: Better width estimation for variable-width fonts.
    static String[] lineBreak (String s, int width) {
	List lineList = new ArrayList();
	int end;
	int start = 0;
	while (start < s.length()) {
	    // First advance to the next non-whitespace character.
	    // Any newlines we come across should produce blank output lines.
	    while (start < s.length()) {
		char c = s.charAt (start);
		if (c == '\n') {
		    lineList.add (" ");
		}
		if (!Character.isWhitespace (c)) {
		    break;
		}
		start++;
	    }

	    // If we are at the end of the string, then we're done.
	    if (start >= s.length()) { break; }

	    // Scan next <width> characters for newline (\n) or end of string.
	    end = Math.min (s.length(), start + width);
	    int p;
	    for (p = start; p < end; p++) {
		char c = s.charAt (p);
		if (c == '\n') {
		    break;
		}
	    }

	    if (end < s.length() && p == end) {
		// If we didn't find a newline, scan backwards until we find whitespace.
		while (end > start) {
		    char c = s.charAt (end);
		    if (Character.isWhitespace (c)) {
			break;
		    }
		    end--;
		}
		// If we didn't find any whitespace, then break apart the word.
		if (end == start) {
		    end = start + width;
		}
	    } else {
		end = p;
	    }

	    // We now have the next line.
	    Util.assert (start < end);
	    Util.assert (start < s.length());
	    Util.assert (end <= s.length());
	    lineList.add (new String (s.substring (start, end)));

	    start = end;
	}
	return (String[]) lineList.toArray (new String [0]);
    }

    static class IntArray {
	int[] values;
	public IntArray () {
	    values = new int [2];
	}
	void grow (int index) {
	    if (index < 0) {
		throw new IndexOutOfBoundsException (
		    "Negative index " + index + "invalid.");
	    }
	    if (index < values.length) {
		return; 
	    }
	    int newLength = values.length;
	    while (index > newLength - 1) {
		newLength *= 2;
	    }
	    int[] newValues = new int [newLength];
	    for (int i = 0; i < values.length; i++) {
		newValues[i] = values[i];
	    }
	    values = newValues;
	}
	public void set (int index, int val) {
	    grow (index);
	    values[index] = val;
	}
	public int get (int index) {
	    grow (index);
	    return values[index];
	}
	public int length () {
	    return values.length;
	}
    }

    static class BoolArray {
	boolean[] values;
	public BoolArray () {
	    values = new boolean [2];
	}
	void grow (int index) {
	    if (index < 0) {
		throw new IndexOutOfBoundsException (
		    "Negative index " + index + "invalid.");
	    }
	    if (index < values.length) {
		return; 
	    }
	    int newLength = values.length;
	    while (index > newLength - 1) {
		newLength *= 2;
	    }
	    boolean[] newValues = new boolean [newLength];
	    for (int i = 0; i < values.length; i++) {
		newValues[i] = values[i];
	    }
	    values = newValues;
	}
	public void set (int index, boolean val) {
	    grow (index);
	    values[index] = val;
	}
	public boolean get (int index) {
	    grow (index);
	    return values[index];
	}
	public int length () {
	    return values.length;
	}
    }

};
